#############################################################################
# tmap_plugin.tcl Copyright 2006, 2007 r0lZ
# PgcEdit plugin to display a summary of the time maps.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "Time Maps" menu will be added in the Plugins menu.
# You need PgcEdit 7.1 or more to use this plugin.
#############################################################################
# Version History:
#
# v 1.3 (March 22, 2007)
# - Fixed a little bug when the PGC duration is EXACTLY one second.  Now, in this case,
#   an empty time map is not considered any more as an error.
# v 1.2 (January 31, 2007)
# - The source DVD path was missing in the log of the Verify Time Maps function.
# v 1.1 (January 22, 2007)
# - bugfix: The Verify Time Maps function crashed if it was called with no DVD loaded in PgcEdit.
# v 1.0 (June 3, 2006)
# - First public release
#############################################################################


# Add Plugins -> Time Maps submenu
# create search and replace submenu
if {! [winfo exists .mainmenu.plugins.tmap]} {
	.mainmenu.plugins add cascade -label "Time Maps" -menu .mainmenu.plugins.tmap
	menu .mainmenu.plugins.tmap
	# add Search & Replace with GUI item
	.mainmenu.plugins.tmap add command -label "Verify Time Maps" -command \
		{::plugins::tmap::verify_tmaps}
	.mainmenu.plugins.tmap add separator
	.mainmenu.plugins.tmap add command -label "About" -command {
			tk_messageBox  -title "Time Maps plugin" -icon info \
					-message "Time Maps plugin for PgcEdit 7.1 or more.\nVersion $::plugins::tmap::version\nAuthor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::tmap {

set version "1.3"

# return the summary of the time maps of the VTS passed in argument
proc verify_tmaps {} {
	if {![info exists ::dvddir]} {bell; return}
	set rebuild_list {}
	set nonseq_warning false

	set txt "Time Map tables\n"
	set txt "${txt}DVD: \"$::dvddir\"\n"
	set txt "${txt}DVD-TEXT General Name: \"[::utils::get_General_Name]\"\n"
	set txt "${txt}Provider ID: \"[::utils::get_provider_ID]\"\n"
	set txt "${txt}Number of VTS: $::pgcs(numvts)"
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set numpgcs $::pgcs($vts,numpgcs)

		set txt "$txt\n\n\n-------------------- VTS $vts --------------------"
		set table $::raw_sectors($vts,VTS_TMAPTI)
		if {$table == ""} {
			set existingtimemaps 0
			set txt "$txt\nNo VTS_TMAPTI table in VTS_[format %02d $vts]_0.IFO!"
		} else {
			set existingtimemaps [get_int table 0 2]
			if {$::pgcs($vts,numpgcs) != $existingtimemaps} {
				set txt "$txt\n********** WARNING: $existingtimemaps VTS_TMAP tables defined in VTS_TMAPTI for $::pgcs($vts,numpgcs) PGCs!"
			} else {
				set txt "$txt\n$existingtimemaps VTS_TMAP tables defined in VTS_TMAPTI for $::pgcs($vts,numpgcs) PGCs."
			}
		}

		for {set pgc 1} {$pgc<=$::pgcs($vts,numpgcs)} {incr pgc} {
			if {$::pgcs($vts,$pgc,numcells) == 0} {continue}
			set seqtitle false
			set pbmode [get_int ::pgcs($vts,$pgc,header) 0xA2 1]
			if {$pbmode == 0} {
				set n 0
				set ttn $::pgcs($vts,$pgc,TTN)
				for {set p 1} {$p<=$::pgcs($vts,numpgcs)} {incr p} {
					if {$::pgcs($vts,$p,TTN) == $ttn} {
						incr n
					}
				}
				if {$n > 1} {
					set mode "non-sequential title"
				} else {
					set mode "sequential title"
					set seqtitle true
				}
			} else {
				if {[expr $pbmode & 0x80]==0} {
					set mode "random PG playback"
				} else {
					set mode "shuffle PG playback"
				}
			}
			set txt "$txt\n\n[::main::pgc_label "" pgc $vts $pgc]  ($mode)"
			if {$table == ""} {
				set txt "$txt\n********** WARNING: No VTS_TMAPTI table in VTS_[format %02d $vts]_0.IFO!"
				lappend rebuild_list [list $vts $pgc "No TMAPTI!"]
			} elseif {$pgc > $existingtimemaps} {
				set txt "$txt\n********** WARNING: No TMAP table defined for PGC $pgc!"
				lappend rebuild_list [list $vts $pgc "No TMAP!"]
			} else {
				set tmapoffset [get_int ::raw_sectors($vts,VTS_TMAPTI) [expr {4+($pgc*4)}] 4]
				set timeunit   [get_int ::raw_sectors($vts,VTS_TMAPTI) $tmapoffset 1]
				set numentries [get_int ::raw_sectors($vts,VTS_TMAPTI) [expr {$tmapoffset+2}] 2]
				set pgctime  [::tmap::BCD2Secs [string range $::pgcs($vts,$pgc,header) 4 7]]
				if {$numentries == 0 || $timeunit == 0} {
					if {$pgctime > 1} {
						if {$seqtitle} {
							set txt "$txt\n********** WARNING: Empty Time Map!"
							lappend rebuild_list [list $vts $pgc "Empty TMAP!"]
						} else {
							set txt "$txt\nEmpty Time Map."
							if {$pgctime > 0} {
								lappend rebuild_list [list $vts $pgc "Empty TMAP (OK: $mode)"]
								if {!$seqtitle} {
									set nonseq_warning true
								}
							}
						}
					} else {
						set txt "$txt\nEmpty Time Map.  (OK: PGC duration too short.)"
					}
				} else {
					set tmaptime [expr $timeunit * $numentries]
					set txt $txt\n[format %-21s "$numentries x $timeunit seconds ="][secs2tc $tmaptime]
					if {[expr $pgctime-$timeunit] > $tmaptime || [expr $pgctime+$timeunit] < $tmaptime} {
						set txt "$txt ; ********** WARNING! **********"
						lappend rebuild_list [list $vts $pgc "Wrong TMAP duration"]
					}
				}
			}
		}
	}
	set txt "$txt\n\n\n------------------- Summary -------------------"
	set nswarn "NOTE: Some of the PGCs with empty time maps are non-sequential PGCs."
	set nswarn "$nswarn\nTheoretically, a non-sequential PGC do not need a time map, but some"
	set nswarn "$nswarn\nnon-compliant players will use the time map anyway.  Therefore, those"
	set nswarn "$nswarn\nPGCs have been added to the list of wrong time maps."
	if {[llength $rebuild_list] == 0} {
		set txt "$txt\nNo Time Map errors detected."
	} else {
		foreach i $rebuild_list {
			set txt "$txt\n[::main::pgc_label "" pgc [lindex $i 0] [lindex $i 1]]: [lindex $i 2]"
		}
		if {$nonseq_warning} {
			set txt "$txt\n\n$nswarn"
		}
	}
	set txt "$txt\n"
	::main::showtext .tmapinfo "Time Map tables" $txt
	if {[llength $rebuild_list] > 0} {
		set msg "There are [llength $rebuild_list] PGCs with missing or wrong time maps!"
		if {$nonseq_warning} {
			set msg "$msg\n\n$nswarn"
		}
		set msg "$msg\n\nDo you want to rebuild the wrong or missing time maps? (recommended)"
		set msg "$msg\n(The time maps will be rebuilt when the DVD is saved.)"
		set r [tk_messageBox -title "Verify Time Maps" -message $msg -type yesno -default yes \
				-icon question]
		if {$r == "yes"} {
			foreach i $rebuild_list {
				::tmap::need_new_tmap [lindex $i 0] [lindex $i 1]
			}
		}
		if {[winfo exists .tmapinfo]} {
			wm deiconify .tmapinfo
		}
	}
	return $rebuild_list
}

proc secs2tc {secs} {
	set hh [expr $secs / 3600]
	set secs [expr $secs % 3600]
	set mm [expr $secs / 60]
	set ss [expr $secs % 60]
	set tm [format "%02d:%02d:%02d" $hh $mm $ss]
	if {[string range $tm 0 2] == "00:"} {
		set tm [string range $tm 3 end]
	}
	if {[string range $tm 0 0] == "0"} {
		set tm [string range $tm 1 end]
	}
	return $tm
}

} ;# end of ::plugins::tmap namespace
